/*
  ALCATEL SpeedTouch USB modem library utility
  Copyright (C) 2001 Benoit PAPILLAULT
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Author   : Benoit PAPILLAULT <benoit.papillault@free.fr>
  Creation : ???

*/

#include "pool.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#define ALLOC_DEBUG

/* for ident(1) command */
static char id[] = "@(#) $Id: pool.c,v 1.2 2001/07/24 01:21:35 edgomez Exp $";

/* this function initialize the pool and allocate the needed memory */
int buffer_pool_init(struct pool *p, int nb, int size)
{
  int i;
  char * ptr;

  /* init data members */
  p->buf_list = NULL;
  p->buf_nb = 0;
  p->big_chunk = NULL;

  /* some checks */
  if ( (p == NULL) || (size <= 0) || (nb <= 0) )
    return 0;

  /* allocate memory for pointers array */
  p->buf_list = (struct buffer *)malloc(sizeof(struct buffer)*nb);
  if (p->buf_list == NULL)
    return 0;

  /* allocate buffers */
  /* done by allocating one big buffer */
  /* better than allocation a lot of small buffers ? */
  p->big_chunk = malloc( nb * size );
  if( p->big_chunk == NULL )
    {
      free(p->buf_list);
      p->buf_list = NULL;

      return 0;
    }

  /* a walking pointer */
  ptr = (char *) p->big_chunk;

  /* init pointer array */
  for (i = 0 ; i < nb ; i++ )
    {
      /* init buffer ptr and mark it as free */
      p->buf_list[i].buf = ptr;
      p->buf_list[i].allocated = 0;
      p->buf_list[i].size = size;

      /* move to the next buffer */
      ptr += size;
    }

  /* init pool structure */
  p->buf_nb = nb;
  p->first_free = 0;

#ifdef ALLOC_DEBUG
  printf("[%p] new pool of %d messages of %d bytes allocated\n",p,p->buf_nb,size);
#endif

  /* init OK ! */
  return 1;
}

/* this function allocate a buffer in the pool */
/* and returns a pointer to it */
/* we start searching in the pool at the entry 'first_free' */
/* which should be free in almost all cases */
int buffer_pool_alloc(struct pool *p, struct buffer ** tbuf)
{
  int i,k;

  for (i = 0 ; i < p->buf_nb ; i++)
    {
      /* calculate entry */
      k = (i + p->first_free) % p->buf_nb;

      if (!p->buf_list[k].allocated)
      {
	p->buf_list[k].allocated = 1;
	*tbuf = &p->buf_list[k];

	/* for the next allocation */
	p->first_free = (k + 1) % p->buf_nb;

#ifdef ALLOC_DEBUG
	printf("[%p] allocated buffer %p (%d) after %d tries\n",p,*tbuf,k,i);
#endif

	return 1;
      }
    }

#ifdef ALLOC_DEBUG
  printf("[%p] no more free buffers !\n",p);
#endif
  return 0;
}

/* free a buffer in the pool */
int buffer_pool_free(struct pool *p, struct buffer * tbuf)
{
  if (!tbuf->allocated)
    {
#ifdef ALLOC_DEBUG
  printf("[%p] buffer %p not allocated !\n",p,tbuf);
#endif
      return 0;
    }

  tbuf->allocated = 0;

#ifdef ALLOC_DEBUG
  printf("[%p] buffer %p freed\n",p,tbuf);
#endif
  return 1;
}

/* search for a designated buffer in the pool */
/* we start searching by the last allocated buffer */
/* and go backward. so we increase the chances to find the buffer early */
/* the buffer is not freed ! */
struct buffer * buffer_pool_search(struct pool *p, void *ptr)
{
  int i,k;

  for (i = p->buf_nb - 1 ; i >= 0; i--)
    {
      /* calculate entry */
      k = (i + p->first_free) % p->buf_nb;

    if (p->buf_list[k].buf == ptr)
      {
#ifdef ALLOC_DEBUG
	printf("[%p] buffer %p found at %d after %d tries\n",p,ptr,k,i);
#endif

	return &p->buf_list[k];
      }
    }

#ifdef ALLOC_DEBUG
  printf("[%p] buffer %p not found\n",p,ptr);
#endif
  return NULL;
}
