/*
  ALCATEL SpeedTouch USB : Simple portable USB interface - *BSD implementation
  Copyright (C) 2001 Benoit PAPILLAULT
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Author   : Richard Tobin
  Creation : May 2001

*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>		/* for strerror */
#include <dev/usb/usb.h>
#include <sys/ioctl.h>

#include "pusb.h"

/* sufficient for most purposes I think... */
#define MAX_CONTROLLERS 10

struct pusb_device_t {
    char *prefix;
    int fd;
    int timeout;
};

struct pusb_endpoint_t {
    int fd;
    int timeout;
};

static pusb_device_t make_device(int fd, const char *prefix)
{
    pusb_device_t dev;

    if(!(dev = malloc(sizeof(*dev))))
	return 0;

    dev->fd = fd;
    dev->timeout = -1;
    if(!(dev->prefix = strdup(prefix)))
    {
	free(dev);
	return 0;
    }

    return dev;
}

static pusb_endpoint_t make_endpoint(int fd)
{
    pusb_endpoint_t ep;

    if(!(ep = malloc(sizeof(*ep))))
	return 0;

    ep->fd = fd;
    ep->timeout = -1;

    return ep;
}

pusb_device_t pusb_search_open(int vendorID, int productID)
{
    int controller, device;
    int cfd, dfd;
    char buf[20], prefix[20];

    for(controller = 0; controller < MAX_CONTROLLERS; controller++)
    {
	sprintf(buf, "/dev/usb%d", controller);
	cfd = open(buf, O_RDONLY);
	if(cfd < 0)
	{
	    if(errno != ENXIO && errno != ENOENT)
		fprintf(stderr, "pusb_search_open: can't open %s: %s\n",
			buf, strerror(errno));
	    continue;
	}

	for(device = 1; device < USB_MAX_DEVICES; device++)
	{
	    struct usb_device_info di;

	    di.addr = device;
	    if(ioctl(cfd, USB_DEVICEINFO, &di) < 0)
		continue;

#ifdef debug
	    {
		int j;
		fprintf(stderr, "something on controller %d device %d\n",
		       controller, device);
		fprintf(stderr, "vendor %s product %s\n", 
			di.vendor, di.product);
		for(j=0; j<4; j++)
		    fprintf(stderr, "name %s\n", di.devnames[j]);
	    }
#endif

	    if(di.vendorNo == vendorID && di.productNo == productID)
	    {
		if(strncmp(di.devnames[0], "ugen", 4) != 0)
		{
		    /* Has a real driver, don't mess with it */
		    continue;
		}

		sprintf(prefix, "/dev/%s", di.devnames[0]);
#if defined(__FreeBSD__)
		sprintf(buf, "%s", prefix);
#else		
		sprintf(buf, "%s.00", prefix);
#endif
		dfd =  open(buf, O_RDWR);

		if(dfd < 0)
		{
		    fprintf(stderr,
			    "pusb_search_open: couldn't open device %s: %s\n",
			    buf, strerror(errno));
		    close(cfd);
		    return 0;
		}
		else
		{
		    close(cfd);
		    return make_device(dfd, prefix);
		}
	    }
	}
    }

#ifdef debug
    fprintf(stderr, "didn't find device\n");
#endif

    close(cfd);
    return 0;
}

pusb_device_t pusb_open(const char *path)
{
    fprintf(stderr, "usb_open: not implemented, exiting\n");
    exit(1);
}

int pusb_close(pusb_device_t dev)
{
    int ret;

    ret = close(dev->fd);
    free(dev->prefix);
    free(dev);

    return ret;
}

int pusb_control_msg(pusb_device_t dev,
		     int request_type, int request,
		     int value, int index, 
		     unsigned char *buf, int size, int timeout)
{
    struct usb_ctl_request req;

    req.request.bmRequestType = request_type;
    req.request.bRequest = request;
    USETW(req.request.wValue, value);
    USETW(req.request.wIndex, index);
    USETW(req.request.wLength, size);

    req.data = buf;
    req.flags = USBD_SHORT_XFER_OK;

    /* !!! If your kernel is built with DIAGNOSTIC (which it is by
       !!! default under some versions of NetBSD) this ioctl will fail.
    */

    if(timeout != dev->timeout)
    {
	if(ioctl(dev->fd, USB_SET_TIMEOUT, &timeout) < 0)
	{
#if defined(__NetBSD__) || defined(__OpenBSD__)
#else
	    return -1;
#endif
	}
	dev->timeout = timeout;
    }

    if(ioctl(dev->fd, USB_DO_REQUEST, &req) < 0)
	return -1;
    else
	return req.actlen;
}

int pusb_set_configuration(pusb_device_t dev, int config)
{
    return ioctl(dev->fd, USB_SET_CONFIG, &config);
}

int pusb_get_configuration(pusb_device_t dev)
{
    int config;

    if(ioctl(dev->fd, USB_GET_CONFIG, &config) < 0)
	return -1;
    return config;
}

int pusb_set_interface(pusb_device_t dev, int interface, int alternate)
{
    struct usb_alt_interface intf;
    int config;

    intf.interface_index = interface;
    intf.alt_no = alternate;

    if(ioctl(dev->fd, USB_SET_ALTINTERFACE, &intf) < 0)
    {
	perror("USB_SET_ALTINTERFACE");
	return -1;
    }

    /* BUG: at least in FreeBSD 4.3, USB_SET_ALTINTERFACE does not
       make the necessary devices for the interface.  So we call
       USB_SET_CONFIG to ensure they are created.  But this will
       fail if any endpoints (other than the control) are open.
       Of course, the devices might already exist, so failure
       to set the configuration does not mean it will not work.
     */
    config = pusb_get_configuration(dev);
    if(config < 0 || pusb_set_configuration(dev, config) < 0)
    {
	fprintf(stderr, "warning: pusb_set_interface may have failed\n");
    }

    return 0;
}

int pusb_get_interface(pusb_device_t dev, int interface)
{
    struct usb_alt_interface intf;

    intf.interface_index = interface;

    if(ioctl(dev->fd, USB_GET_ALTINTERFACE, &intf) < 0)
	return -1;

    return intf.alt_no;
}

pusb_endpoint_t pusb_endpoint_open(pusb_device_t dev, int epnum, int flags)
{
    int one = 1;
    char buf[20];
    int fd;

#if defined(__FreeBSD__)
    sprintf(buf, "%s.%d", dev->prefix, epnum & 0x0f);
#else
    sprintf(buf, "%s.%02d", dev->prefix, epnum & 0x0f);
#endif
    fd = open(buf, flags, 0);
    if(fd < 0)
    {
	fprintf(stderr,
		"pusb_endpoint_open: couldn't open endpoint %s: %s\n",
		buf, strerror(errno));
	return 0;
    }

    if(ioctl(fd, USB_SET_SHORT_XFER, &one) < 0)
	return 0;

    return make_endpoint(fd);
}

int pusb_endpoint_read(pusb_endpoint_t ep, 
		       unsigned char *buf, int size, int timeout)
{
    int ret;

    if(timeout != ep->timeout)
    {
	if(ioctl(ep->fd, USB_SET_TIMEOUT, &timeout) < 0)
	    return -1;
	ep->timeout = timeout;
    }

    ret = read(ep->fd, buf, size);
#ifdef debug
	fprintf(stderr, "pusb_endpoint_read: read returned %d\n", ret);
#endif

	return ret;
}

int pusb_endpoint_write(pusb_endpoint_t ep, 
			const unsigned char *buf, int size, int timeout)
{
    int sent = 0;
    int ret;

    if(timeout != ep->timeout)
    {
	if(ioctl(ep->fd, USB_SET_TIMEOUT, &timeout) < 0)
	    return -1;
	ep->timeout = timeout;
    }

    do {
	ret = write(ep->fd, buf+sent, size-sent);
	if(ret < 0)
	    return sent > 0 ? sent : -1;
	sent += ret;
    } while(ret > 0 && sent < size);

    return sent;
}

int pusb_endpoint_close(pusb_endpoint_t ep)
{
    int ret;

    ret = close(ep->fd);
    free(ep);

    return ret;
}

int pusb_claim_interface(pusb_device_t dev,int interface)
{
    return 0;
}

int pusb_release_interface(pusb_device_t dev,int interface)
{
    return 0;
}
